from dataclasses import dataclass
from typing import Optional

import pandas as pd

from hawa.common.query import MetaUnit
from hawa.config import project
from hawa.paper.health import HealthApiData


@dataclass
class ClassMixin:
    """为了在 __mro__ 中有更高的优先级， mixin 在继承时，应该放在最前"""
    meta_unit_type: str = 'class'


@dataclass
class ClassHealthApiData(ClassMixin, HealthApiData):
    """使用 school id 作为 meta_unit_id，额外增加班级的逻辑"""
    meta_class_id: Optional[int] = None  # 必填
    class_id: Optional[int] = None
    class_name: Optional[str] = ''

    def _to_init_a0_meta(self):
        if not self.grade:
            raise ValueError("grade 必填")
        if not self.meta_class_id:
            raise ValueError("meta_class_id 必填")
        self.class_id = self.meta_class_id % 100
        self.class_name = f"{self.grade}年级{self.class_id}班"

    def _to_init_a_meta_unit(self):
        super()._to_init_a_meta_unit()
        self.meta_unit = MetaUnit(
            id=self.meta_unit_id, name=f"{self.meta_unit.name}{self.class_name}",
            short_name=self.meta_unit.short_name
        )

    def _to_init_e_answers(self):
        """筛选班级的答案"""
        super()._to_init_e_answers()
        records = []
        for _, row in self.answers.iterrows():
            if int(str(row['student_id'])[:15]) == self.meta_class_id:
                records.append(row)
        self.answers = pd.DataFrame.from_records(records)
        project.logger.debug(f'class answers: {len(self.answers)}')



"""通用的 report data 构造器，支持 校、区、市、省、全国级别的通用报告数据构造"""

import json
from collections import Counter, defaultdict
from copy import deepcopy
from dataclasses import dataclass, field
from decimal import Decimal, ROUND_HALF_UP
from typing import Optional, ClassVar, Any, Set

import pandas as pd
import pendulum

from hawa.base.db import DbUtil, RedisUtil
from hawa.base.decos import log_func_time
from hawa.base.errors import NoCasesError, NoValidAnswers
from hawa.common.query import DataQuery
from hawa.common.utils import GradeData, CaseData, Measurement, Util
from hawa.config import project


class MetaCommonData(type):
    def __new__(cls, name, bases, attrs):
        attrs['db'] = DbUtil()
        attrs['redis'] = RedisUtil()
        attrs['query'] = DataQuery()
        return super().__new__(cls, name, bases, attrs)


@dataclass
class TargetsCount:
    targets: set[str] = field(default_factory=set)
    count: int = 0
    field_extra: str = ''
    field_extra_count: int = 0


@dataclass
class CommonData(metaclass=MetaCommonData):
    # 构造单位
    meta_unit_type: Optional[str] = ''  # class/school/group/district/city/province/country
    meta_unit_id: Optional[int] = None
    meta_unit: Optional[Any] = None
    grade: Optional[int] = None  # 必填

    different_mode: Optional[str] = 'default'  # 用于区分数据的不同模式 default默认/xx新乡

    # 时间目标
    target_year: Optional[int] = None
    last_year_num: Optional[int] = None
    is_load_last: bool = True  # 仅计算往年数据时 为 False

    is_load_all: bool = True  # 加载全部数据
    is_filter_cls_less10: bool = True

    # 卷子
    test_type: str = ''
    test_types: list = field(default_factory=list)
    code_word_list: Set[str] = field(default=set)  # 卷子使用指标的词表，详见继承

    # meta class tool
    db: ClassVar[DbUtil] = None
    redis: ClassVar[RedisUtil] = None
    query: ClassVar[DataQuery] = None

    # 原始数据

    school_ids: list[int] = field(default_factory=list)
    schools: pd.DataFrame = field(default_factory=pd.DataFrame)

    codebook: pd.DataFrame = field(default_factory=pd.DataFrame)

    cases: pd.DataFrame = field(default_factory=pd.DataFrame)
    case_ids: list[int] = field(default_factory=list)
    case_project_ids: Counter = field(default_factory=Counter)

    answers: pd.DataFrame = field(default_factory=pd.DataFrame)
    item_codes: pd.DataFrame = field(default_factory=pd.DataFrame)

    students: pd.DataFrame = field(default_factory=pd.DataFrame)
    student_ids: list[int] = field(default_factory=list)
    student_count: Optional[int] = None

    item_ids: Optional[set[int]] = None
    items: Optional[pd.DataFrame] = None

    # 辅助工具
    grade_util: Optional[GradeData] = None
    case_util: Optional[CaseData] = None
    rank_names = ['待提高', '中等', '良好', '优秀']
    measurement = Measurement()

    # 计算数据
    final_answers: pd.DataFrame = field(default_factory=pd.DataFrame)
    final_scores: pd.DataFrame = field(default_factory=pd.DataFrame)
    grade_final_scores: pd.DataFrame = field(default_factory=pd.DataFrame)

    # 去年全国数据
    last_year = None
    last_year_code_scores: Optional[pd.DataFrame] = field(default_factory=pd.DataFrame)

    def __post_init__(self):
        # 初始化数据
        if self.is_load_all:
            self.load_all_data()

            # 构建辅助工具
            self._to_build_helper()

            # 计算数据
            count_functions = [i for i in dir(self) if i.startswith('_to_count_')]
            for func in count_functions:
                getattr(self, func)()
                project.logger.info(f'count {func} success')

        else:
            self.load_less_data()
            self._to_build_helper()

    def _to_init_a_meta_unit(self):
        try:
            self.meta_unit = self.query.query_unit(self.meta_unit_type, str(self.meta_unit_id))
        except TypeError as e:
            project.logger.warning(f'query_unit error: {e}')
            self.__class__.query = DataQuery()
            self.meta_unit = self.query.query_unit(self.meta_unit_type, str(self.meta_unit_id))
        if '百万' in self.meta_unit.name:
            self.different_mode = 'xx'

    def _to_init_b_time(self):
        if not self.target_year:
            self.target_year = pendulum.now().year
        self.last_year_num = self.target_year - 1
        project.logger.info(f'target_year: {self.target_year}')

    def _to_init_c_schools(self):
        if self.school_ids:
            self.schools = self.query.query_schools_by_ids(self.school_ids)
        else:
            match self.meta_unit_type:
                case 'country':
                    self.schools = self.query.query_schools_all()
                case 'province':
                    self.schools = self.query.query_schools_by_startwith(self.meta_unit_id // 10000)
                case 'city':
                    self.schools = self.query.query_schools_by_startwith(self.meta_unit_id // 100)
                case 'district':
                    self.schools = self.query.query_schools_by_startwith(self.meta_unit_id)
                case 'school' | 'class' | 'student':
                    self.schools = self.query.query_schools_by_ids([self.meta_unit_id])
                case 'group':
                    self.schools = self.query.query_schools_by_group_id(group_id=self.meta_unit_id)
                case _:
                    raise ValueError(f'unknown meta_unit_type: {self.meta_unit_type}')
            self.school_ids = self.schools['id'].tolist()
        project.logger.debug(f'schools: {len(self.schools)}')

    def _to_init_d_cases(self, is_cleared: bool = True):
        start_stamp = pendulum.datetime(self.target_year, 1, 1)
        end_stamp = pendulum.datetime(self.target_year + 1, 1, 1)
        start_stamp_str = start_stamp.format(project.format)
        end_stamp_str = end_stamp.format(project.format)

        papers = self.query.query_papers(test_types=self.test_types, test_type=self.test_type)
        paper_ids = papers['id'].tolist()

        self.cases = self.query.query_cases(
            school_ids=self.school_ids,
            paper_ids=paper_ids,
            valid_to_start=start_stamp_str,
            valid_to_end=end_stamp_str,
            is_cleared=is_cleared
        )
        if self.cases.empty:
            raise NoCasesError(f'no cases:{self.meta_unit} {self.school_ids}')
        self.case_ids = self.cases['id'].tolist()
        self.school_ids = self.cases['school_id'].unique().tolist()
        project.logger.debug(f'cases: {len(self.cases)}')

        if self.grade:
            self.cases = self.cases.loc[self.cases['id'] % 100 == self.grade, :]
            self.case_ids = self.cases['id'].tolist()
            project.logger.debug(f'cases: {len(self.cases)}')
        if len(self.cases) == 0:
            raise NoCasesError(f'grade {self.grade} cases is empty')

    @log_func_time
    def _to_init_e_answers(self):
        self.answers = self.query.query_answers(case_ids=self.case_ids)
        project.logger.debug(f'answers: {len(self.answers)}')
        if self.answers.empty:
            raise NoValidAnswers(f"{self.meta_unit.id} no valid answers")

    def _to_init_f_students(self):
        self.student_ids = set(self.answers['student_id'].tolist())
        student_id_list = list(self.student_ids)
        self.students = self.query.query_students(student_id_list, mode=self.different_mode)
        self.student_count = len(self.students)
        project.logger.debug(f'students: {self.student_count}')

    def _to_init_g_items(self):
        self.item_ids = set(self.answers['item_id'].drop_duplicates())
        self.items = self.query.query_items(self.item_ids)
        project.logger.debug(f'items: {len(self.items)}')

    def _to_init_y_item_codes(self):
        word_list = self.code_word_list | {'other'} if len(self.code_word_list) == 1 else self.code_word_list
        self.item_codes = self.query.query_item_codes(self.item_ids, categories=word_list)

    def _to_init_z_dim_field(self):
        cache_key = f"{project.PROJECT}:codebook"
        if data := self.redis.conn.get(cache_key):
            self.codebook = pd.DataFrame.from_records(json.loads(data))
        else:
            self.codebook = self.query.query_codebook()
            cache_data = self.codebook.to_json(orient='records', force_ascii=False)
            self.redis.conn.set(cache_key, cache_data, ex=60 * 60 * 24 * 7)

    def _to_build_helper(self):
        self.grade = GradeData(case_ids=self.case_ids)
        self.case = CaseData(cases=self.cases)
        self.grade_util = GradeData(case_ids=self.case_ids)
        self.case_util = CaseData(cases=self.cases)

    @log_func_time
    def _to_count_a_final_answers(self):
        items = {k: {} for k in self.code_word_list}
        # code-dimension/field  ~  item_id  ~ code   name
        project.logger.debug(f"{self.code_word_list=}")
        for (item_id, category), codes in self.item_codes.groupby(['item_id', 'category']):
            if category in self.code_word_list:
                for _, code_data in codes.iterrows():
                    items[category][item_id] = code_data['name']

        data = pd.merge(
            self.answers, self.students.loc[:, ['id', 'gender', 'nickname']],
            left_on='student_id', right_on='id'
        )
        project.logger.debug(f"ans merge students {len(data)}")
        # inner 时，final_answers 和 answers 数目不等：final_answers 过滤掉了 没有 code_word_list（维度领域或其他）的题目
        # outer 时，数目相等，不过滤任何题目
        data = pd.merge(data, self.item_codes, left_on='item_id', right_on='item_id', how='inner')

        project.logger.debug(f'merge success {data.shape}')

        self.set_data_extra(d=data, set_grade=True, set_class=True)

        data['username'] = data['nickname']
        for code_word in self.code_word_list:
            data[code_word] = data.item_id.apply(
                lambda x: self.get_code_name_items(item_id=x, word=code_word, items=items))
        self.final_answers = data.drop_duplicates(subset=['case_id', 'student_id', 'item_id'])

        if self.is_filter_cls_less10:
            self.final_answers = self.filter_answers_cls_less10(ans=self.final_answers)
            self.student_ids = set(self.final_answers['student_id'].tolist())

        project.logger.debug(f'final_answers: {len(self.final_answers)}')

    @log_func_time
    def _to_count_b_final_scores(self):
        self.final_scores = self.count_final_score(answers=self.final_answers)
        project.logger.debug(f'final_scores: {len(self.final_scores)}')

    def get_code_name_items(self, item_id, word: str, items: dict):
        try:
            return items[word][item_id]
        except KeyError:
            return ''

    @staticmethod
    def count_level(score, mode: str = 'f'):
        assert mode in ('f', 'r'), 'only support feedback or report'
        if score >= 90:
            a = 'A'
        elif score >= 80:
            a = 'B'
        elif score >= 60:
            a = 'C'
        else:
            a = 'D'
        key = "RANK_LABEL" if mode == 'r' else 'FEEDBACK_LEVEL'
        return project.ranks[key][a]

    def count_final_score(self, answers: pd.DataFrame):
        records = []
        for student_id, group in answers.groupby('student_id'):
            score = group.score.mean() * 100
            record = {
                "student_id": student_id,
                "username": self.get_col_value(group['username']),
                "grade": self.get_col_value(group['grade']),
                "gender": self.get_col_value(group['gender']),
                "score": score,
                "level": self.count_level(score),
                "cls": self.get_col_value(group['cls']),
            }
            records.append(record)
        return pd.DataFrame.from_records(records)

    def get_last_year_miss(self, grade: int):
        key = f'{project.REDIS_PREFIX}{self.last_year_num}:data'
        data = json.loads(self.redis.conn.get(key))
        try:
            grade_data = data[str(grade)]
        except KeyError:
            for i in range(1, grade):
                temp_grade = grade - i
                grade_data = data.get(str(temp_grade))
                if grade_data:
                    break
            else:
                grade_data = data[str(3)]

        grade_data['people']['grade'] = f"{grade}年级"
        for k, v in grade_data['code'].items():
            grade_data['code'][k]['grade'] = grade
        return grade_data

    def sort_rank(self, name: str) -> int:
        order_map = dict(zip(self.rank_names, range(0, 4)))
        return order_map[name]

    def count_dim_field_ranks(self, item_code: str):
        """
        计算维度、领域的 ranks 分级比例
        :param item_code:dimision or field
        """
        r = defaultdict(list)
        codes = set()
        for (s, c), student_code_group in self.final_answers.groupby(['student_id', item_code]):
            s_c_score = Util.format_num(student_code_group.score.mean() * 100, project.precision)
            codes.add(c)
            r[c].append(s_c_score)
        codes = list(codes)
        df = pd.DataFrame.from_records(r)
        res = {}
        for c in codes:
            row = df[c]
            base_row_ranks = {k: 0 for k in self.rank_names}
            count_row_ranks = pd.cut(
                row, bins=[0, 60, 80, 90, 100], labels=self.rank_names,
                right=False, include_lowest=True,
            ).value_counts().to_dict()
            sum_value = sum(count_row_ranks.values())
            row_ranks = {k: Util.format_num(v / sum_value * 100, project.precision) for k, v in
                         (base_row_ranks | count_row_ranks).items()}
            res[c] = row_ranks
        code_map = self.get_dim_field_order(key=item_code)
        return {
            "data": res, "codes": sorted(codes, key=lambda x: code_map[x]),
            "legend": self.rank_names,
        }

    def count_sub_units(self, target_level: str = 'school'):
        """查询下辖单位"""
        match target_level:
            case 'school':
                return self.school_ids
            case 'district':
                return {i // (10 ** 4) for i in self.school_ids}
            case _:
                raise ValueError(f"target_level: {target_level} not support")

    def get_cascade_students(self):
        """年级/班级/学生嵌套"""
        data = self.final_scores
        self.set_data_extra(d=data, set_grade=False, set_class=True)
        res = []
        for grade, grade_group in data.groupby('grade'):
            grade_row = {
                'label': f'{grade}年级', 'value': grade,
                'children': [], "is_leaf": False
            }
            for cls, cls_group in grade_group.groupby('cls'):
                cls_row = {
                    'label': f'{cls}班', 'value': int(cls),
                    'children': [], "is_leaf": False
                }
                for _, student_g_row in cls_group.iterrows():
                    student_row = {
                        'label': student_g_row['username'],
                        'value': str(student_g_row['student_id']),
                        "is_leaf": True
                    }
                    cls_row['children'].append(student_row)
                if not cls_row['children']:
                    cls_row['is_leaf'] = True
                grade_row['children'].append(cls_row)
            if not grade_row['children']:
                grade_row['is_leaf'] = True
            res.append(grade_row)
        return res

    def get_cascade_schools_from_province(self):
        """省-市-区县-学校的 cascade 数据"""
        sch_ids = self.school_ids
        province_ids = {i // (10 ** 8) * 10000 for i in sch_ids}
        city_ids = {i // (10 ** 6) * 100 for i in sch_ids}
        district_ids = {i // (10 ** 4) for i in sch_ids}
        location_ids = province_ids | city_ids | district_ids
        locations = self.query.query_locations(list(location_ids))
        schools = self.schools
        location_map = {lo['id']: lo for _, lo in locations.iterrows()}
        school_map = {sch['id']: sch for _, sch in schools.iterrows()}
        res = []
        for p_id in province_ids:
            p = location_map[p_id]
            p_row = {
                'label': p['name'], 'value': p['id'], 'children': [], "is_leaf": False
            }
            if p_id not in project.municipality_ids:
                for c_id in city_ids:
                    if c_id // 10000 * 10000 != p_id:
                        continue
                    c = location_map[c_id]
                    c_row = {
                        'label': c['name'], 'value': c['id'], 'children': [], "is_leaf": False
                    }
                    for d_id in district_ids:
                        if d_id // 100 * 100 != c_id:
                            continue
                        d = location_map[d_id]
                        d_row = {
                            'label': d['name'], 'value': d['id'], 'children': [], "is_leaf": False
                        }
                        for s_id in sch_ids:
                            if s_id // 10000 != d_id:
                                continue
                            s = school_map[s_id]
                            s_row = {
                                'label': s['name'], 'value': s['id'], "is_leaf": True
                            }
                            d_row['children'].append(s_row)
                        c_row['children'].append(d_row)
                    p_row['children'].append(c_row)
                res.append(p_row)
            else:
                for d_id in district_ids:
                    if d_id // 10000 * 10000 != p_id:
                        continue
                    d = location_map[d_id]
                    d_row = {
                        'label': d['name'], 'value': d['id'], 'children': [], "is_leaf": False
                    }
                    for s_id in sch_ids:
                        if s_id // 10000 != d_id:
                            continue
                        s = school_map[s_id]
                        s_row = {
                            'label': s['name'], 'value': s['id'], "is_leaf": True
                        }
                        d_row['children'].append(s_row)
                    p_row['children'].append(d_row)
                res.append(p_row)
        return res

    def get_dim_field_order(self, key: str):
        """获取维度/领域顺序的映射
        :param key: dimension/field
        """
        data = self.codebook.loc[self.codebook['category'] == key, :]
        order_map = {i['name']: i['order'] for _, i in data.iterrows()}
        if '其他' in order_map.keys():
            del order_map['其他']
        return order_map

    def load_less_data(self):
        init_functions = [i for i in dir(self) if i.startswith('_to_init_')]
        for func in init_functions:
            if '_to_init_e_' in func:
                break
            getattr(self, func)()

    def load_all_data(self):
        init_functions = [i for i in dir(self) if i.startswith('_to_init_')]
        for func in init_functions:
            getattr(self, func)()
            project.logger.info(f'load data {func} success')

    def set_data_extra(self, d: pd.DataFrame, set_grade: bool = True, set_class: bool = True):
        """为 数据 设置不同模式下的 extra 字段"""
        if self.different_mode == 'xx':
            user_classes = {}
            user_grades = {}
            for _, row in self.students.iterrows():
                extra = json.loads(row['extra'])
                user_grades[row['id']] = extra.get('grade', 0)
                user_classes[row['id']] = extra.get('class', 0)
            if set_grade:
                d['grade'] = d['student_id'].apply(lambda x: user_grades.get(x, ''))
            if set_class:
                d['cls'] = d['student_id'].apply(lambda x: user_classes.get(x, ''))
        else:
            if set_grade:
                d['grade'] = d['case_id'].apply(lambda x: x % 100)
            if set_class:
                d['cls'] = d['student_id'].apply(lambda x: int(str(x)[13:15]))

    @property
    def psy_item_ids(self):
        items = self.query.query_phq_items()
        return set(items['id'].tolist())

    @property
    def mgarbage_item_ids(self):
        items = self.query.query_mgarbage_items()
        return set(items['id'].tolist())

    def count_rank_students(self):
        """计算各等级人数"""
        base_ranks = {k: 0 for k in project.ranks['FEEDBACK_LEVEL'].values()}
        count_ranks = self.final_scores['level'].value_counts().to_dict()
        return base_ranks | count_ranks

    @staticmethod
    def get_col_value(col):
        return col.tolist()[0]

    def filter_answers_cls_less10(self, ans: pd.DataFrame):
        """过滤班级学生数小于10的班级"""

        # 步骤2: 对student_id去重，然后根据grade和cls分组，计数每组的学生数量
        grouped_data = ans.drop_duplicates(subset='student_id').groupby(['grade', 'cls']).size().reset_index(
            name='student_count')

        # 步骤3: 过滤出学生人数大于或等于10的年级和班级
        filtered_groups = grouped_data[grouped_data['student_count'] >= 10]

        # 步骤4: 使用过滤后的数据过滤原始数据集
        filtered_data = ans[
            ans.set_index(['grade', 'cls']).index.isin(filtered_groups.set_index(['grade', 'cls']).index)]

        # 显示过滤后的数据
        return filtered_data

    def count_dim_or_field_scores_by_answers(self, answers, item_code, res_format: str = 'dict'):
        """
        计算维度或领域得分
        :param answers: 由 final answers 中取出的部分数据，属于某一主体
        :param item_code: dimension/field
        :param res_format: dict/list
        :return:
        """
        keys, values, mapping = [], [], {}
        for code, code_group in answers.groupby(item_code):
            score = Util.format_num(code_group.score.mean() * 100, project.precision)
            keys.append(code)
            values.append(score)
            mapping[code] = score
        match res_format:
            case 'dict':
                return mapping
            case 'list':
                code_map = self.get_dim_field_order(key=item_code)
                keys = sorted(keys, key=lambda x: code_map[x])
                values = [mapping[k] for k in keys]
                return keys, values

    def get_grade_focus(self, limit: int = 60, step: int = 5):
        """获取所有年级的优先关注点"""
        res = {}
        for grade, grade_answers in self.final_answers.groupby('grade'):
            dimensions = self.count_dim_or_field_scores_by_answers(
                answers=grade_answers, item_code='dimension', res_format='dict'
            )
            fields = self.count_dim_or_field_scores_by_answers(
                answers=grade_answers, item_code='field', res_format='dict'
            )
            grade_res = self.get_limit_focus_recu_res(dimensions=dimensions, fields=fields, limit=limit, step=step)
            res[grade] = grade_res
        return res

    def get_limit_focus_recu_res(self, dimensions: dict, fields: dict, limit: int, step: int = 2):
        res = [k for k, v in (dimensions | fields).items() if v < limit]
        if res:
            return res
        if not res:
            return self.get_limit_focus_recu_res(dimensions, fields, limit + step, step)
        return res

    @staticmethod
    def count_mean_score_by_final_scores(scores: pd.DataFrame):
        return round(scores.score.mean(), 1)

    def count_11scores_by_answers(self, ans: pd.DataFrame):
        """通过传入的 answers 计算 11 个得分 6f 4d 1total"""
        res = {
            "total": self.count_mean_score_by_final_scores(scores=self.count_final_score(answers=ans))
        }
        for f, field_ans in ans.groupby('field'):
            field_score = self.count_mean_score_by_final_scores(scores=self.count_final_score(answers=field_ans))
            res[f] = field_score
        for dim, dim_ans in ans.groupby('dimension'):
            dim_score = self.count_mean_score_by_final_scores(scores=self.count_final_score(answers=dim_ans))
            res[dim] = dim_score
        return res

    def count_gender_scores33(self, ans: pd.DataFrame, grade: Optional[int] = None, cls: Optional[int] = None):
        """计算 total/M/F 11 共 33 个得分"""
        res = {
            "total": self.count_11scores_by_answers(ans=ans) | {
                'gender': 'total', 'grade': grade, 'cls': cls,
                'group': self.count_group_name(cls=cls, gender='total')
            }
        }
        for gender, gender_ans in ans.groupby('gender'):
            res[gender] = self.count_11scores_by_answers(gender_ans) | {
                'gender': gender, 'grade': grade, 'cls': cls,
                'group': self.count_group_name(cls=cls, gender=str(gender))
            }
        return res

    @staticmethod
    def count_group_name(cls: int, gender: str = 'total'):
        """计算 11 33 组别名称"""
        prefix = f'{cls}班' if cls else ''
        suffix_map = {'total': '全体', 'M': '男生', 'F': '女生'}
        return f'{prefix}{suffix_map[gender]}'

    def count_class_detail_scores11(self):
        """
        计算学校的全部年级、班级维度领域得分 每单位 11 3组 共33个
        列：年级 组别 班级（无/年级 有/具体班级）性别（total/M/F） 6领域列 4维度列 总分列
        """
        rows33 = []
        for grade, grade_ans in self.final_answers.groupby('grade'):
            grade_gender_scores = self.count_gender_scores33(ans=grade_ans, grade=grade, cls=None)
            rows33.append(grade_gender_scores)
            for cls, grade_cls_ans in grade_ans.groupby('cls'):
                grade_cls_gender_scores = self.count_gender_scores33(ans=grade_cls_ans, grade=grade, cls=cls)
                rows33.append(grade_cls_gender_scores)
        base_res = []
        for row33 in rows33:
            for k, v in row33.items():
                base_res.append(v)
        return base_res

    @staticmethod
    def _count_point_name(row, point_map):
        return point_map.get(row['code'].rsplit('.', 1)[0].replace('target1', 'point'), '')

    @staticmethod
    def _count_field_name(row, field_map):
        return field_map.get(row['code'].rsplit('.', 2)[0].replace("target1", "domain"), '')

    @staticmethod
    def _retain_prec(n: float, prec: int = 1):
        n = n * 100
        return int(n) if n in (0, 0.0, 100.0, 100) else round(n, prec)

    @staticmethod
    def _count_reverse_sorted_rank(rank_data: dict):
        base = [(v, k) for k, v in rank_data.items()]
        res = [b for b in sorted(base, key=lambda x: x[0], reverse=True)]
        return res

    def count_rank_dis_by_final_scores(self, scores: pd.DataFrame):
        """计算 scores 的 rank 分布比例"""
        base = dict(
            zip(project.ranks['FEEDBACK_LEVEL'].values(),
                [0] * len(project.ranks['FEEDBACK_LEVEL'])))
        count = base | scores.level.value_counts().to_dict()
        count = {k: v / sum(count.values()) for k, v in count.items()}
        return {k: self._retain_prec(v) for k, v in count.items()}

    def count_field_point_target(self, page_limit: int = 38):
        periods = tuple(self.grade_util.grade_periods + ['-'])
        codes_sql = f"select * from code_guide where period in {periods} or category in ('G.domain','G.point');"
        codes = self.query.raw_query(codes_sql)
        targets = codes.loc[codes['category'] == 'G.target1', :]
        targets['target'] = targets['name']
        target1_points = codes.loc[codes['category'] == 'G.point', :]
        target1_points_map = {i['code']: i['name'] for _, i in target1_points.iterrows()}
        fields = codes.loc[codes['category'] == 'G.domain', :]
        fields_map = {i['code']: i['name'] for _, i in fields.iterrows()}
        targets['point'] = targets.apply(lambda x: self._count_point_name(x, target1_points_map), axis=1)
        targets['field'] = targets.apply(lambda x: self._count_field_name(x, fields_map), axis=1)
        field_count_map = targets['field'].value_counts().to_dict()
        point_count_map = targets['point'].value_counts().to_dict()
        targets['field_count'] = targets['field'].apply(lambda x: field_count_map.get(x, 0))
        targets['point_count'] = targets['point'].apply(lambda x: point_count_map.get(x, 0))
        targets['field_prefix'] = targets['code'].apply(lambda x: x.split('.')[2])
        targets['point_prefix'] = targets['code'].apply(lambda x: f"{x.split('.')[2]}.{x.split('.')[3]}")
        cols = ['field', 'point', 'target', 'field_count', 'point_count', 'field_prefix', 'point_prefix', 'field_extra']

        new_target_counts = {row['target']: row['field_count'] for _, row in targets.iterrows()}
        new_target_field_extra = {row['target']: row['field'] for _, row in targets.iterrows()}
        target_counts = []
        for the_field, field_group in targets.groupby('field'):
            field_count = field_group['field_count'].values[0]
            if field_count < page_limit:
                continue

            the_target_count = TargetsCount(field_extra=the_field)

            for point_prefix, point_group in field_group.groupby('point_prefix'):
                point_count = point_group['point_count'].values[0]

                # 单段终止条件
                if the_target_count.count + point_count > page_limit:
                    target_counts.append(the_target_count)
                    new_extra_count = the_target_count.field_extra_count + 1
                    the_target_count = TargetsCount(
                        field_extra=f'{the_field}extra{new_extra_count}', field_extra_count=new_extra_count)

                the_target_count.count += point_count
                the_target_count.targets |= set(point_group['target'].tolist())
            else:
                target_counts.append(the_target_count)

        for the_tc in target_counts:
            new_target_counts |= {k: the_tc.count for k in the_tc.targets}
            new_target_field_extra |= {k: the_tc.field_extra for k in the_tc.targets}

        targets['field_count'] = targets.apply(lambda x: new_target_counts.get(x['target'], 0), axis=1)
        targets['field_extra'] = targets.apply(lambda x: new_target_field_extra.get(x['target'], ''), axis=1)

        res = targets.loc[:, cols]
        return res.to_dict(orient='records')

    @staticmethod
    def count_rank_by_score(score: float):
        if score >= 90:
            level = 'A'
        elif score >= 80:
            level = 'B'
        elif score >= 60:
            level = 'C'
        else:
            level = 'D'
        return project.ranks['FEEDBACK_LEVEL'][level]

    @staticmethod
    def _count_item_targets(percent_item_id_map, all_item_codes: pd.DataFrame, all_code_guides: pd.DataFrame):
        code_category = 'G.target1'
        res = []
        for percent, item_ids in percent_item_id_map.items():
            conditions = (all_item_codes['item_id'].isin(item_ids) &
                          (all_item_codes['category'] == code_category))
            item_targets = all_item_codes.loc[conditions, :]
            merge_targets = all_code_guides.merge(item_targets, on='code', how='inner', suffixes=('', '_y'))
            merge_targets['percent'] = percent
            record = {
                "name": '，'.join(merge_targets['name'].unique().tolist()),
                "percent": percent
            }
            res.append(record)
        df = pd.DataFrame.from_records(res)
        return df

    @staticmethod
    def _count_top_last_item_ids(item_scores: pd.Series, key: str = 'top'):
        res = defaultdict(list)
        score_filter = set()
        func_map = {
            "top": item_scores.nlargest,
            "last": item_scores.nsmallest
        }
        for k, v in func_map[key](n=len(item_scores)).items():
            decimal_number = Decimal(v).quantize(Decimal('0.001'), rounding=ROUND_HALF_UP) * 100
            if len(score_filter) < 3:
                score_filter.add(decimal_number)
                res[decimal_number].append(k)
            else:
                break
        return res

    def count_grade_class_item_target(self):
        """计算各年级各班级 top3/last3 item target 相对优势/优先关注点"""
        all_item_codes = self.query.query_item_codes(item_ids=list(self.item_ids), categories=None)
        all_code_guides = self.query.query_code_guides()
        res = {}
        for grade, grade_ans in self.final_answers.groupby('grade'):
            grade_data = []
            for cls, grade_cls_ans in grade_ans.groupby('cls'):
                cls_score = self.count_mean_score_by_final_scores(scores=self.count_final_score(answers=grade_cls_ans))
                cls_rank = self.count_rank_by_score(score=cls_score)
                item_scores = grade_cls_ans.groupby('item_id').score.mean()
                self._count_top_last_item_ids(item_scores=item_scores)
                # get top3/last3 item_ids
                top3_item_ids = self._count_top_last_item_ids(item_scores=item_scores, key='top')
                last3_item_ids = self._count_top_last_item_ids(item_scores=item_scores, key='last')

                top3_targets = self._count_item_targets(
                    percent_item_id_map=top3_item_ids,
                    all_item_codes=all_item_codes, all_code_guides=all_code_guides
                )
                last3_targets = self._count_item_targets(
                    percent_item_id_map=last3_item_ids,
                    all_item_codes=all_item_codes, all_code_guides=all_code_guides
                )
                dimensions = self.count_dim_or_field_scores_by_answers(
                    answers=grade_cls_ans, item_code='dimension', res_format='dict'
                )
                fields = self.count_dim_or_field_scores_by_answers(
                    answers=grade_cls_ans, item_code='field', res_format='dict'
                )
                upper_codes, lower_codes = [], []
                for k, v in (dimensions | fields).items():
                    if v >= 80:
                        upper_codes.append(k)
                    elif v <= 60:
                        lower_codes.append(k)

                top3_names = top3_targets['name'].unique().tolist()
                last3_names = last3_targets['name'].unique().tolist()
                record = {
                    "cls": cls, "grade": grade, "score": cls_score, "rank": cls_rank,
                    "top3": top3_names, "last3": last3_names,
                    "top3_data": top3_targets.to_dict(orient='records'),
                    "last3_data": last3_targets.to_dict(orient='records'),
                    "upper_codes": upper_codes, "lower_codes": lower_codes
                }
                grade_data.append(record)
            res[grade] = grade_data
        return res

    @staticmethod
    def count_str_rank(rank: dict):
        """"""
        rank = deepcopy(rank)
        rank['素养'] = round(rank['优秀'] + rank['良好'], 1)
        rank['基础'] = round(rank['待提高'] + rank['中等'], 1)
        res = {k: str(v) for k, v in rank.items()}
        return res

    @property
    def grade_class_map(self):
        """生成年级/班级映射"""
        res = {}
        max_min_score_class = defaultdict(list)
        for grade, grade_ans in self.final_answers.groupby('grade'):
            res[grade] = {"cls": []}
            max_cls, min_cls = grade_ans['cls'].max(), grade_ans['cls'].min()
            for cls, cls_ans in grade_ans.groupby('cls'):
                student_scores = self.count_final_score(answers=cls_ans)
                gc_score = self.count_mean_score_by_final_scores(scores=student_scores)
                max_min_score_class[grade].append((cls, gc_score))
                cls_student_count = len(cls_ans['student_id'].unique())
                boy_ans = cls_ans.loc[cls_ans['gender'] == 'M', :]
                girl_ans = cls_ans.loc[cls_ans['gender'] == 'F', :]
                cls_boy_count = len(boy_ans['student_id'].unique())
                cls_girl_count = len(girl_ans['student_id'].unique())
                cls_rank = self.count_rank_dis_by_final_scores(scores=student_scores)
                cls_reverse_rank = self._count_reverse_sorted_rank(cls_rank)
                cls_11_scores = self.count_11scores_by_answers(cls_ans)
                del cls_11_scores['total']
                reverse_dim_field_scores = self._count_reverse_sorted_rank(cls_11_scores)
                cls_record = {
                    "cls": cls, "student_count": cls_student_count,
                    "boy_count": cls_boy_count, "girl_count": cls_girl_count,
                    "avg_score": gc_score, "max_score": round(student_scores['score'].max(), 1),
                    "boy_score": self.count_mean_score_by_final_scores(scores=self.count_final_score(boy_ans)),
                    "girl_score": self.count_mean_score_by_final_scores(scores=self.count_final_score(girl_ans)),

                    "min_score": round(student_scores['score'].min(), 1),
                    "rank": cls_rank,
                    "str_rank": {},
                    "boy_rank": self.count_rank_dis_by_final_scores(scores=self.count_final_score(answers=boy_ans)),
                    "girl_rank": self.count_rank_dis_by_final_scores(scores=self.count_final_score(answers=girl_ans)),
                    "reverse_rank": cls_reverse_rank, "reverse_dim_fields": reverse_dim_field_scores,
                    "gender_codes_score": self.count_gender_scores33(ans=cls_ans),
                }
                for k in ('rank', 'boy_rank', 'girl_rank'):
                    cls_record['str_rank'][k] = self.count_str_rank(rank=cls_record[k])
                res[grade]["cls"].append(cls_record)
            sort_cls_scores = sorted(max_min_score_class[grade], key=lambda x: x[1])
            res[grade] |= {
                # 班级顺序的最大最小值
                "max_cls": max_cls, "min_cls": min_cls,
                # 得分最高最低的班级
                "high": sort_cls_scores[-1][0], "low": sort_cls_scores[0][0]
            }
        return res



import json
from collections import defaultdict
from dataclasses import dataclass
from typing import Optional

from hawa.base.errors import NoAnswersError
from hawa.common.query import DataQuery
from hawa.config import project
from hawa.paper.health import HealthApiData
from hawa.paper.mht import MhtPlusApiData


@dataclass
class StudentMixin:
    """为了在 __mro__ 中有更高的优先级， mixin 在继承时，应该放在最前"""
    meta_unit_type: str = 'student'
    meta_student_id: Optional[int] = None  # 必填
    student_name: Optional[str] = ''

    def _to_init_a0_meta(self):
        if not self.meta_student_id:
            raise ValueError("meta_student_id 必填")


@dataclass
class StudentHealthApiData(StudentMixin, HealthApiData):
    is_filter_cls_less10: bool = False

    def _to_init_a_meta_unit(self):
        try:
            self.meta_unit = self.query.query_unit(self.meta_unit_type, str(self.meta_student_id))
        except TypeError as e:
            project.logger.warning(f'query_unit error: {e}')
            self.__class__.query = DataQuery()
            self.meta_unit = self.query.query_unit(self.meta_unit_type, str(self.meta_unit_id))
        self.student_name = self.meta_unit.name
        if self.meta_unit.client_id == 10:
            self.different_mode = 'xx'

    def _to_init_d_cases(self, is_cleared: bool = True):
        super()._to_init_d_cases(is_cleared=False)

    def _to_init_e_answers(self):
        """筛选学生的答案"""
        self.answers = self.query.query_answers(case_ids=self.case_ids, student_id=self.meta_student_id)

        if len(self.answers) == 0:
            raise NoAnswersError(f"学生 {self.meta_student_id} 没有答题记录")

        project.logger.debug(f'student answers: {len(self.answers)}')


def count_segment_scores(score: float, limit_group: list[float]) -> list[float]:
    reduce_limit = [item - limit_group[i - 1] if i else item for i, item in enumerate(limit_group)]
    for i, limit in enumerate(limit_group):
        if score <= limit:
            if i == 0:
                return [score]
            else:
                return reduce_limit[:i] + [score - limit_group[i - 1]]


def get_mht_code_describe(code: str, level: str):
    mht_map = {
        "学习焦虑": {
            "非常健康": "非常健康(3分以下)：学习焦虑低，学习不会受到困扰，能正确对待考试成绩。",
            "健康": "正常波动（3-8分）：学习焦虑在正常范围波动，建议主动关注自己的学习、考试等相关问题，在必要的时候主动寻求教师、家长和同学帮助。",
            "预警": "单项预警(8分以上)：对考试怀有恐惧心理，无法安心学习，十分关心考试分数。这类被试必须接受为他制定的有针对性的特别指导计划。"
        },
        "对人焦虑": {
            "非常健康": "非常健康(3分以下)：热情，大方，容易结交朋友。",
            "健康": "正常波动（3-8分）：对人焦虑在正常范围波动，建议主动关注自己的人际适应、师生关系、亲子关系、同学关系等相关问题，在必要的时候主动寻求教师、家长和同学帮助。",
            "预警": "单项预警(8分以上)：过分注重自己的形象，害怕与人交往，退缩。这类被试必须接受为他制定的有针对性的特别指导计划。"
        },
        "孤独倾向": {
            "非常健康": "非常健康(3分以下)：爱好社交，喜欢寻求刺激，喜欢与他人在一起。",
            "健康": "正常波动（3-8分）：孤独倾向在正常范围波动，建议主动关注自己心境和情绪的波动，在必要的时候主动寻求教师、家长和同学帮助。",
            "预警": "单项预警(8分以上)：孤独、抑郁，不善与人交往，自我封闭。这类被试必须接受为他制定的有针对性的特别指导计划。"
        },
        "自责倾向": {
            "非常健康": "非常健康(3分以下)：自信，能正确看待失败。",
            "健康": "正常波动（3-8分）：自责倾向在正常范围波动，建议主动关注自己对成败、挫折的认知，在必要的时候主动寻求教师、家长和同学帮助。",
            "预警": "单项预警(8分以上)：自卑，常怀疑自己的能力，常将失败、过失归咎于自己。这类被试必须接受为他制定的有针对性的特别指导计划。"
        },
        "过敏倾向": {
            "非常健康": "非常健康(3分以下)：敏感性较低，能较好地处理日常事物。",
            "健康": "正常波动（3-8分）：过敏倾向在正常范围波动，建议主动关注自己对学习、生活和人际关系的调整需求，在必要的时候主动寻求教师、家长和同学帮助。",
            "预警": "单项预警(8分以上)：过于敏感，容易为一些小事而烦恼。这类被试必须接受为他制定的有针对性的特别指导计划。"
        },
        "身体症状": {
            "非常健康": "非常健康(3分以下)：基本没有身体异常表现。",
            "健康": "正常波动（3-8分）：身体症状在正常范围波动，建议主动关注自己身体生理上的明显改变，在必要的时候主动寻求教师、家长和同学帮助。",
            "预警": "单项预警(8分以上)：在极度焦虑的时候，会出现呕吐失眠、小便失禁等明显症状。这类被试必须接受为他制定的有针对性的特别指导计划。"
        },
        "恐怖倾向": {
            "非常健康": "非常健康(3分以下)：基本没有恐怖感。",
            "健康": "正常波动（3-8分）：恐怖倾向在正常范围波动，建议主动关注自己对日常生活中常见事物的恐怖反应，在必要的时候主动寻求教师、家长和同学帮助。",
            "预警": "单项预警(8分以上)：对某些日常事物，如黑暗等，有较严重的恐惧感。这类被试必须接受为他制定的有针对性的特别指导计划。"
        },
        "冲动倾向": {
            "非常健康": "非常健康(3分以下)：基本没有冲动。",
            "健康": "正常波动（3-8分）：冲动倾向在正常范围波动，建议主动关注自己对情况、行为的冲动倾向，在必要的时候主动寻求教师、家长和同学帮助。",
            "预警": "单项预警(8分以上)：十分冲动，自制力差。这类被试必须接受为他制定的有针对性的特别指导计划。"
        }
    }
    return mht_map[code][level]


@dataclass
class StudentMhtPlusApiData(StudentMixin, MhtPlusApiData):
    """"""
    meta_student_id: Optional[int] = None  # 必填
    student_name: Optional[str] = ''
    is_filter_cls_less10: bool = False

    def _to_init_a_meta_unit(self):
        try:
            self.meta_unit = self.query.query_unit(self.meta_unit_type, str(self.meta_student_id))
        except TypeError as e:
            project.logger.warning(f'query_unit error: {e}')
            self.__class__.query = DataQuery()
            self.meta_unit = self.query.query_unit(self.meta_unit_type, str(self.meta_unit_id))
        if self.meta_unit.client_id == 10:
            self.different_mode = 'xx'
        self.student_name = self.meta_unit.name

    def _to_init_d_cases(self, is_cleared: bool = True):
        super()._to_init_d_cases(is_cleared=False)

    def _to_init_e_answers(self):
        """筛选学生的答案"""
        self.answers = self.query.query_answers(case_ids=self.case_ids, student_id=self.meta_student_id)

        if len(self.answers) == 0:
            raise NoAnswersError(f"学生 {self.meta_student_id} 没有答题记录")

        project.logger.debug(f'student answers: {len(self.answers)}')

    def count_student_archive(self):
        """获取学生档案"""
        student = self.students.iloc[0]
        school = self.schools.iloc[0]
        extra = json.loads(student.extra)

        # source 2021浙江心理 9 source mht 100
        items = self.items
        answers = self.answers
        if len(answers) not in (109, 129):
            raise ValueError('答题数量不对')
        answer_map = {a.item_id: a.answer for _, a in answers.iterrows()}
        item_codes = self.item_codes.loc[self.item_codes.category == 'mht', :]
        codes = self.codebook.loc[self.codebook.code.isin(item_codes.code), :]
        item_chinese_code_names = {c['code']: c['name'] for _, c in codes.iterrows()}
        item_code_map = {row['item_id']: item_chinese_code_names[row['code']] for _, row in item_codes.iterrows()}

        records, ph9_records, mht_records = [], [], []
        for _, item in items.iterrows():
            try:
                record = {
                    'item_id': item['id'], 'answer': answer_map[item.id], 'code': item_code_map.get(item['id'], ''),
                    "score": 0, 'category': item.source
                }
            except KeyError:
                continue
            if item.source == 'mht' and answer_map[item.id] == 'A':
                record['score'] = 1
            elif item.source == '2021浙江心理':
                ans_score_map = dict(zip('ABCDE', range(5)))
                record['score'] = ans_score_map[answer_map[item.id]]
            match item.source:
                case 'mht':
                    mht_records.append(record)
                case '2021浙江心理':
                    ph9_records.append(record)
            records.append(record)

        ph9_score = sum([i['score'] for i in ph9_records])
        ph9_descrite = '没有忧郁症(注意自我保重)'
        if ph9_score >= 20:
            ph9_descrite = '可能有重度忧郁症 (一定要看心理医生或精神科医生)'
        elif ph9_score >= 15:
            ph9_descrite = '可能有中重度忧郁症 (建议咨询心理医生或精神科医生)'
        elif ph9_score >= 10:
            ph9_descrite = '可能有中度忧郁症 (最好咨询心理医生或心理医学工作者)'
        elif ph9_score >= 5:
            ph9_descrite = '可能有轻微忧郁症 (建议咨询心理医生或心理医学工作者)'

        mht_score = sum([i['score'] for i in mht_records if i['code'] != '效度'])

        mht_thing_score = defaultdict(float)
        mht_group_records = defaultdict(list)
        mht_radar_codes, mht_radar_score = [], []
        for mht_r in mht_records:
            if mht_r['code'] != '效度':
                mht_thing_score[mht_r['code']] += mht_r['score']
                mht_group_records[mht_r['code']].append(mht_r)

        is_super_bad = False
        if max(mht_thing_score.values()) > 10:
            is_super_bad = True
        for k, v in mht_thing_score.items():
            mht_radar_codes.append({"name": k, "max": 15})
            mht_radar_score.append(v)

        mht_code_records = []
        for code, mht_group in mht_group_records.items():
            code_score = mht_thing_score[code]
            code_describe = '非常健康'
            if code_score >= 8:
                code_describe = '预警'
            elif code_score >= 3:
                code_describe = '健康'

            record = {
                "code": code, "max": 15, "total_score": code_score,
                "describe": code_describe,
                "scores": count_segment_scores(score=code_score, limit_group=[3, 8, 15]),
                "detail": get_mht_code_describe(code=code, level=code_describe)
            }
            mht_code_records.append(record)
        return {
            "student": {
                "school": school['name'],
                "gender": student.gender,
                "name": student.nickname,
                "grade": extra['grade'] if extra.get('grade') else extra['case_id'][-2:],
                "klass": extra.get('class', extra.get('klass', extra.get('cls', 0)))
            },
            "score": {
                "ph9": ph9_score,
                "ph9_scores": count_segment_scores(score=ph9_score, limit_group=[4, 9, 14, 19, 27]),
                "mht": mht_score,
                "mht_code_records": mht_code_records,
                "radar": {
                    "codes": mht_radar_codes,
                    "scores": mht_radar_score
                }
            },
            "result": {
                "ph9": ph9_descrite,
            },
            "is_super_bad": is_super_bad,
        }



"""仅为集合，无上级单位"""
from dataclasses import dataclass

from hawa.base.errors import NoSchoolIdsError
from hawa.common.query import MetaUnit
from hawa.paper.health import HealthApiData
from hawa.paper.mht import MhtPlusQusApiData, MhtPlusApiData


@dataclass
class AssembleMixin:
    """
    为了在 __mro__ 中有更高的优先级， mixin 在继承时，应该放在最前
    学校集合
    """
    meta_unit_id: int = 0
    meta_unit_type: str = 'assemble'


@dataclass
class AssembleHealthApiData(AssembleMixin, HealthApiData):
    """依赖 school_ids 工作"""

    def _to_init_a_meta_unit(self):
        self.meta_unit = MetaUnit(id=self.meta_unit_id, name='学校集合', short_name='学校集合')

    def _to_init_c_schools(self):
        if not self.school_ids:
            raise NoSchoolIdsError("Assemble 参数 school_Ids 不能为空")
        super()._to_init_c_schools()


@dataclass
class AssembleMhtPlusPlusApiData(AssembleMixin, MhtPlusApiData):
    """依赖 school_ids 工作"""


@dataclass
class AssembleMhtPlusQusPlusApiData(AssembleMixin, MhtPlusQusApiData):
    """依赖 school_ids 工作"""



"""健康测评数据"""
import itertools
import json
import math
from collections import defaultdict
from dataclasses import dataclass, field
from pprint import pprint
from typing import Union, Set

import pandas as pd
from munch import Munch
from pingouin import cronbach_alpha

from hawa.common.data import CommonData
from hawa.common.utils import Util
from hawa.config import project


@dataclass
class HealthData(CommonData):
    """健康测评数据，不应直接使用，应向下继承 city/district/school 等"""
    test_types: list[str] = field(default_factory=lambda: ['publicWelfare', 'ZjpublicWelfare', 'XxPublicWelfareQus'])
    code_word_list: Set[str] = field(default_factory=lambda: {'dimension', 'field'})

    def replace_hai(self, text: str, condition: list):
        if condition:
            return text
        else:
            return text.replace('还', '')


@dataclass
class HealthApiData(HealthData):
    """为 yingde api 项目提供基类"""

    def score_rank(self, grade: int, gender: str = ''):
        """
        某年级 健康素养水平各等级占比
        :param grade: 年级
        :param gender: 性别 '' 全部 'M' 男 'F' 女
        """
        base = {'优秀': 0, '良好': 0, '中等': 0, '待提高': 0}
        fs = final_scores = self.__get_grade_final_scores(grade=grade)
        if gender:
            final_scores = fs.loc[fs.gender == gender, :]
        raw = final_scores.level.value_counts().to_dict()
        data = base | raw
        sum_data = sum(data.values())
        if not sum_data:
            raise ValueError(f'grade {grade} score rank is empty')
        percent = {k: Util.format_num(v * 100 / sum_data, project.precision) for k, v in data.items()}
        return percent

    def gender_compare(self, grade: int):
        """某年级学生健康素养水平及性别比较图"""
        final_scores = self.__get_grade_final_scores(grade=grade)
        total_score = final_scores.score.mean()
        m_score = final_scores.loc[final_scores.gender == 'M'].score.mean()
        f_score = final_scores.loc[final_scores.gender == 'F'].score.mean()
        return {
            "total": Util.format_num(total_score, project.precision),
            "M": Util.format_num(m_score, project.precision),
            "F": Util.format_num(f_score, project.precision),
        }

    def dim_field_gender_compare(self, grade: int, item_code: str, key_format: str = 'en'):
        """某年级六大领域/四大维度测评得分及性别比较图"""
        final_answers = self.__get_grade_final_answers(grade=grade)
        answers = final_answers.loc[~final_answers[item_code].isnull(), :]
        raw_data = {
            'total': answers,
            'M': answers.loc[answers.gender == 'M', :],
            'F': answers.loc[answers.gender == 'F', :]
        }
        res = {}
        for gender_k, gender_v in raw_data.items():
            res[gender_k] = self.count_dim_or_field_scores_by_answers(
                answers=gender_v, item_code=item_code, res_format='dict'
            )
        map_order = self.get_dim_field_order(key=item_code)
        codes = answers[item_code].unique().tolist()
        codes = sorted(codes, key=lambda x: map_order[x])
        gender_box = []
        for gender_k, gender_data in res.items():
            row_data = {
                "name": gender_k if key_format == 'en' else project.total_gender_map[gender_k],
                "value": []
            }
            for c in codes:
                row_data["value"].append(res[gender_k][c])
            gender_box.append(row_data)

        return {
            "category": codes,
            "values": gender_box,
        }

    def count_student_grade(self, student_id: int):
        """计算指定学生的年级"""
        fa = self.final_answers
        student_answers = fa.loc[fa['student_id'] == int(student_id), :]
        case_ids = student_answers['case_id'].unique().tolist()
        return case_ids[0] % 100

    def __get_grade_final_answers(self, grade: int):
        return self.final_answers[self.final_answers['grade'] == grade]

    def __get_grade_final_scores(self, grade: int):
        return self.final_scores[self.final_scores['grade'] == grade]

    def get_class_scores(self):
        """获取年级各班级的分数 仅可在 school/assemble 中使用"""
        scores = self.final_scores
        scores['cls'] = scores['student_id'].apply(lambda x: f"{int(str(x)[13:15])}班")
        res = scores.groupby('cls').score.mean().to_dict()
        sorted_keys = sorted(res.keys(), key=lambda x: int(x[:-1]))
        sorted_res = {k: res[k] for k in sorted_keys}
        keys, values = sorted_res.keys(), [Util.format_num(i) for i in sorted_res.values()]
        return {'keys': list(keys), 'values': list(values)}


@dataclass
class HealthReportData(HealthData):
    # 计算数据
    code_scores: pd.DataFrame = field(default_factory=pd.DataFrame)
    summary_scores: dict = field(default_factory=dict)
    grade_good_bad = None  # 优势 优先关注点

    grade_rank_dis = None  # 年级、性别、水平学生分布占比
    grade_str_rank_dis = None  # 年级、性别、水平学生分布占比
    grade_reverse_rank_dis = None  # 占比由高到低的年级分布水平
    grade_gender_distribution = None
    grade_score = None  # 年级最高、最低、平均分
    grade_code_score = None  # 健康素养水平 所需数据
    compare_grade_year_school = None  # 比较学校与去年维度领域
    compare_all_total = None  # 比照去年数据和学校数据的全年级平均数

    def _to_count_a0_last_year_data(self):
        if not self.is_load_last:
            return
        key = f'{project.REDIS_PREFIX}{self.last_year_num}:data'
        if not self.redis.conn.exists(key):
            raise ValueError(f'last year data not exists: {key}')
        self.last_year = json.loads(self.redis.conn.get(key))
        res = []
        for grade in self.grade.grades:
            grade_data = self.get_last_year_miss(grade=grade)
            year_code_score = pd.DataFrame(grade_data['code']).T
            year_code_score['grade'] = int(grade)
            year_code_score['i'] = [int(grade) * 10 + i for i in range(1, 11)]
            year_code_score.set_index('i', inplace=True)
            res.append(year_code_score)
        self.last_year_code_scores = pd.concat(res)

    def _to_count_c_code_scores(self):
        records = []
        for grade in self.grade.grades:
            for code in self.measurement.dimensions:
                record = self._count_grade_code_score(grade=grade, code=code, category='dimension')
                records.append(record)
            for code in self.measurement.fields:
                record = self._count_grade_code_score(grade=grade, code=code, category='field')
                records.append(record)
        self.code_scores = pd.DataFrame.from_records(records)

    def _to_count_d_summary_scores(self):
        records = defaultdict(dict)
        for grade, group in self.code_scores.groupby('grade'):
            for gender in project.gender_map.keys():
                records[grade][gender] = Util.format_num(self._count_school_score(group, key=gender), project.precision)
        self.summary_scores = records

    def _to_count_e_grade_good_bad(self):
        records = defaultdict(dict)
        for g, group in self.code_scores.groupby('grade'):
            for gender in project.gender_map.keys():
                records[g][gender] = {
                    'dimension': {'good': [], 'bad': []},
                    'field': {'good': [], 'bad': []}
                }
            for _, row in group.iterrows():
                for gender in project.gender_map.keys():
                    if row[gender] >= 80:
                        if row['category'] == 'dimension':
                            records[g][gender]['dimension']['good'].append(row['code'])
                        if row['category'] == 'field':
                            records[g][gender]['field']['good'].append(row['code'])
                    elif row[gender] < 60:
                        if row['category'] == 'dimension':
                            records[g][gender]['dimension']['bad'].append(row['code'])
                        if row['category'] == 'field':
                            records[g][gender]['field']['bad'].append(row['code'])
                    else:
                        pass
                else:
                    for code in itertools.chain(self.measurement.dimensions, self.measurement.fields):
                        for ccc in ['dimension', 'field']:
                            if code in records[g]['total'][ccc]['bad']:
                                for ggg in 'MF':
                                    try:
                                        records[g][ggg][ccc]['bad'].remove(code)
                                    except ValueError:
                                        pass

        self.grade_good_bad = records

    def _to_count_f_case_gender_counts(self):
        """单年级，各班级性别数据；多年级，各年级性别数据。"""
        ans = self.final_answers.drop_duplicates(subset=['student_id'])
        data = ans.loc[:, ['grade', 'cls', 'gender', 'student_id']]
        # 年级 班级 男生数 女生数 总人数
        # 班级 男生数 女生数 总人数
        records = []
        for grade, grade_group in data.groupby(by='grade'):
            for cls, cls_group in grade_group.groupby('cls'):
                grade_cls = Munch(grade=grade, cls=cls)
                cls_total = Munch(total=len(cls_group))
                cls_gender = cls_group.gender.value_counts().to_dict()
                records.append(grade_cls | cls_total | cls_gender)
            grade_total = Munch(total=len(grade_group))
            grade_gender = grade_group.gender.value_counts().to_dict()
            records.append({'grade': grade, 'cls': 0, 'F': 0, 'M': 0} | grade_total | grade_gender)
        records = pd.DataFrame.from_records(records).fillna(0)
        records['M'] = records.M.astype('int')
        records['F'] = records.F.astype('int')
        records['grade'] = records.grade.apply(lambda x: f"{project.grade_simple[x]}年级")
        records = records.loc[records.cls == 0, :]
        self.case_gender_counts = records.to_dict(orient='records')

    def _to_count_g_cronbach_alpha(self):
        cols = ['student_id', 'item_id', 'score']
        res = []
        for grade, group in self.final_answers.groupby('grade'):
            base = group.loc[:, cols]
            data = pd.pivot_table(base, index='item_id', columns='student_id', values='score')
            c: tuple = cronbach_alpha(data)
            res.append(c[0])
        self.cronbach_alpha = [round(i, 3) for i in res]

    def _to_count_h_grade_gender_distribution(self):
        """年级性别分布"""
        data = self.case_gender_counts
        records = {}
        for row in data:
            records[row['grade']] = Munch(row)
        self.grade_gender_distribution = records

    def _to_count_i_grade_score(self):
        """年级最高、最低、平均分"""
        records = Munch()
        for grade in self.grade.grades:
            ans = self.final_answers.loc[self.final_answers.grade == grade, :]
            score = ans.groupby('student_id').score
            record = Munch(
                avg=self._retain_prec(score.mean().mean()),
                min=self._retain_prec(score.mean().min()),
                max=self._retain_prec(score.mean().max()),
                rank=self.count_rank_by_score(score.mean().mean()*100)
            )
            records[grade] = record
        self.grade_score = records

    def _to_count_j_grade_rank_dis(self):
        records = {}
        str_records = {}
        for grade, group in self.final_scores.groupby('grade'):
            count = self.count_rank_dis_by_final_scores(scores=group)
            records[grade] = Munch()
            str_records[grade] = Munch()
            for gender, g in group.groupby('gender'):
                gender_count = self.count_rank_dis_by_final_scores(scores=g)
                records[grade][gender] = gender_count
                str_records[grade][gender] = self.count_str_rank(rank=gender_count)
            records[grade].total = count
            str_records[grade].total = self.count_str_rank(rank=count)

        self.grade_rank_dis = records
        self.grade_str_rank_dis = str_records

    def _to_count_k_grade_reverse_rank_dis(self):
        records = {}
        rank_dis = self.grade_rank_dis
        for g in self.grade.grades:
            try:
                here = rank_dis[g].total
            except KeyError:
                continue
            records[g] = self._count_reverse_sorted_rank(rank_data=here)
        self.grade_reverse_rank_dis = records

    def _to_count_l_grade_code_score(self):
        """健康素养水平 所需数据"""
        records = {}
        scores = self.code_scores
        for grade in self.grade.grades:
            base: pd.DataFrame = scores.loc[scores.grade == grade, :]
            base_dimensions = base.loc[base.category == 'dimension', ['code', 'total']]
            base_fields = base.loc[base.category == 'field', ['code', 'total']]

            dimensions = self._count_df_reverse(first_col='code', second_col='total', data=base_dimensions)
            fields = self._count_df_reverse(first_col='code', second_col='total', data=base_fields)
            a = base.loc[base.category == 'field', ['F']].mean().mean()
            b = base.loc[base.category == 'field', ['M']].mean().mean()
            a = 0 if math.isnan(a) else a
            b = 0 if math.isnan(b) else b
            records[grade] = {
                'dimension': dimensions, 'field': fields,
                'cond': self.count_cond(a, b, target='男生')
            }
        self.grade_code_score = records

    def _to_count_n_compare_grade_year_school(self):
        if not self.is_load_last:
            return
        res = defaultdict(dict)
        cols = ['total', 'M', 'F']
        for grade in self.grade.grades:
            for col in cols:
                res[grade][col] = self._count_dim_field_diff(grade=grade, key=col)
        self.compare_grade_year_school = res

    def _to_count_o_compare_all_total(self):
        if not self.is_load_last:
            return
        data = []
        for grade in self.grade.grades:
            sch = self.grade_score[grade].avg
            last_year_grade_data = self.get_last_year_miss(grade=grade)
            year = last_year_grade_data['score']['total']
            data.append(sch - year)
        self.compare_all_total = self.count_cond(a=sum(data) / len(data), b=0)

    @classmethod
    def cache_year_data(cls, year: int):
        """缓存年数据"""
        project.logger.info(f"缓存年数据 {year=}")
        key = f'{project.REDIS_PREFIX}{year}:data'

        res = defaultdict(dict)
        tool = cls(meta_unit_type='country', meta_unit_id=0, target_year=year, last_year_num=year - 1,
                   is_load_last=False)
        project.logger.debug('after init data')

        for grade, value in zip(tool.grade.grades, tool.grade_gender_distribution.values()):
            res[grade]['people'] = value
        for grade, group in tool.code_scores.groupby('grade'):
            res[grade]['code'] = {}
            for r in group.to_dict(orient='records'):
                res[grade]['code'][r['code']] = r
        for grade, value in tool.grade_rank_dis.items():
            res[grade]['rank'] = value
        score_dict = defaultdict(dict)
        # 计算年级、性别 平均分
        for grade, first_value in res.items():
            for second_key, second_value in first_value.items():
                if second_key == 'code':
                    total_score = cls._count_year_score('total', second_value)
                    f_score = cls._count_year_score('F', second_value)
                    m_score = cls._count_year_score('M', second_value)
                    score_dict[grade] = {
                        'total': total_score, 'F': f_score, 'M': m_score
                    }
        for grade, value in score_dict.items():
            res[grade]['score'] = value
        # 计算总人数
        res['total'] = defaultdict(dict)
        res['total']['total'] = sum([cls._get_value(v, 'total') for v in res.values()])
        res['total']['M'] = sum([cls._get_value(v, 'M') for v in res.values()])
        res['total']['F'] = sum([cls._get_value(v, 'F') for v in res.values()])
        cls.redis.conn.set(key, json.dumps(res, ensure_ascii=False))

    def _count_grade_code_score(self, code: str, grade: int, category: str):
        """计算指定年级、指定维度、领域的分数"""
        ans = self.final_answers
        local_ans = ans.loc[(ans.grade == grade) & (ans[category] == code), :]
        student_score = local_ans.groupby('student_id').score.mean().mean() * 100
        gender_score = local_ans.groupby(['gender', 'student_id']) \
                           .score.mean().reset_index().groupby(
            'gender').score.mean() * 100
        res = Munch(
            total=student_score, grade=grade, code=code, category=category,
            **gender_score.to_dict())
        return res

    @staticmethod
    def _count_school_score(group, key: str):
        """计算学校的维度、领域性别分数"""
        dim_score = group.loc[group.category == 'dimension', :][key].mean()
        field_score = group.loc[group.category == 'field', :][key].mean()
        return (dim_score + field_score) / 2

    def _count_cls(self, row):
        if row['cls']:
            return f"{project.grade_simple[row['grade']]}({int(row['cls'])})班"
        else:
            return f"{project.grade_simple[row['grade']]}年级"

    @staticmethod
    def _get_value(v, k):
        """计算去年数据中的总人数"""
        base: Union[dict, int] = dict(v).get('people', 0)
        if base:
            return base[k]
        else:
            return 0

    @staticmethod
    def _count_year_score(category: str, data: dict):
        """
        :param category: total/M/F
        :param data: 需计算数据
        """
        dim_score = sum([dict(v)[category] for v in list(data.values()) if v['category'] == 'dimension']) / 4
        field_score = sum([dict(v)[category] for v in list(data.values()) if v['category'] == 'field']) / 6
        score = (dim_score + field_score) / 2
        return score

    @property
    def gender_count(self) -> Munch:
        r = self.students.gender.value_counts()
        return Munch({'M': 0, 'F': 0} | r.to_dict())

    # 工具函数

    def _count_df_reverse(self, first_col: str, second_col: str, data: pd.DataFrame):
        base = data.to_dict(orient='records')
        middle = {d[first_col]: d[second_col] / 100 for d in base}
        reverse_middle = {v: k for k, v in middle.items()}
        res = [(self._retain_prec(k), reverse_middle[k]) for k in sorted(reverse_middle.keys(), reverse=True)]
        return res

    def count_cond(self, a: float, b: float, target: str = ''):
        a = 0 if math.isnan(a) else a
        b = 0 if math.isnan(b) else b
        if a == b:
            condition = f'等于{target}' if target else '等于'
        elif a - b >= 5:
            condition = f'明显高于{target}' if target else '明显高于'
        elif abs(a - b) < 5:
            condition = f'与{target}的差异不明显' if target else '差异不明显于'
        elif a - b <= -5:
            condition = f'明显低于{target}' if target else '明显低于'
        else:
            raise
        return condition

    def _count_dim_field_diff(self, grade: int, key: str):
        """计算学校与全国对比的维度、领域高低
        :param key: 比较的项：total/M/F
        """
        data = self._build_one_data(grade=grade)
        key_col = f'{key}_y_s'
        data[key_col] = data[f"{key}_sch"] - data[f"{key}_year"]
        cols = ['category_sch', 'code', key_col, f'{key}_year', f'{key}_sch']
        compare = data.sort_values(['category_sch', key_col]).loc[:, cols]
        dim_compare = compare.loc[compare.category_sch == 'dimension', :]
        field_compare = compare.loc[compare.category_sch == 'field', :]
        dim_low, dim_high = dim_compare.head(n=1), dim_compare.tail(n=1)
        field_low, field_high = field_compare.head(n=1), field_compare.tail(n=1)
        return {
            'dim_low': dim_low.code.values[0] if dim_low[key_col].values[0] < 0 else '',
            'field_low': field_low.code.values[0] if field_low[key_col].values[0] < 0 else '',
            'dim_high': dim_high.code.values[0] if dim_high[key_col].values[0] > 0 else '',
            'field_high': field_high.code.values[0] if field_high[key_col].values[0] > 0 else '',
        }

    def _build_one_data(self, grade: int):
        """抽象某个构建数据的逻辑"""
        year_scores = self.last_year_code_scores
        sch_scores = self.code_scores
        year_code_score = year_scores.loc[year_scores.grade == grade, :]
        school_code_score = sch_scores.loc[sch_scores.grade == grade, :]
        data = pd.merge(
            year_code_score, school_code_score, left_on='code', right_on='code',
            suffixes=['_year', '_sch']
        )
        return data

    def build_code_gender_data(self, category: str, g: int):
        data = self._build_one_data(grade=g)
        use_cols = [f'{category}_year', f"{category}_sch", 'code', 'category_sch']
        data = data.loc[:, use_cols].sort_values(['category_sch', 'code'])
        del data['category_sch']
        return data

    def compare_gender_text(self, grade: int):
        grade_rank_dis_m = self.grade_rank_dis[grade].M
        grade_rank_dis_f = self.grade_rank_dis[grade].F
        level_m = Util.format_num(sum([grade_rank_dis_m['优秀'], grade_rank_dis_m['良好']]), project.precision)
        level_f = Util.format_num(sum([grade_rank_dis_f['优秀'], grade_rank_dis_f['良好']]), project.precision)
        if abs(level_m - level_f) >= 5:
            if level_m > level_f:
                return '男生明显高于女生'
            else:
                return '男生明显低于女生'
        elif abs(level_m - level_f) >= 3:
            if level_m > level_f:
                return '男生略高于女生'
            else:
                return '男生略低于女生'
        else:
            return '男生与女生不存在明显差异'

    def low_high_code_text(self, grade: int, gender: str):
        """相对最高、相对最低文本
        :param grade: total/M/F
        :param gender: total/M/F
        """
        if not self.is_load_last:
            return
        data = self.build_code_gender_data(category=gender, g=grade)
        data['diff'] = data[f"{gender}_sch"] - data[f"{gender}_year"]
        bigger = data.loc[data['diff'] >= 5, :].code.to_list()
        smaller = data.loc[data['diff'] <= -5, :].code.to_list()
        res = ''
        if bigger:
            local_codes = self._build_codes(bigger)
            res += f"{local_codes}{project.category_map[gender]}分数" \
                   f"明显高于全国{project.grade_simple[grade]}年级{project.category_map[gender]}平均分数，"
        if smaller:
            local_codes = self._build_codes(smaller)
            res += f"{local_codes}{project.category_map[gender]}分数" \
                   f"明显低于全国{project.grade_simple[grade]}年级{project.category_map[gender]}平均分数，"
        if not bigger and not smaller:
            res += '所有维度和领域都没有明显差异。'
        else:
            res += '其余维度和领域没有明显差异。'
        return res

    def _build_codes(self, codes: list[str]):
        if len(codes) == 1:
            return f'“{codes[0]}”'
        elif len(codes) > 1:
            code_text = '、'.join([f'“{i}”' for i in codes[:-1]]) + f'和“{codes[-1]}”'
            return f"在{code_text}维度和领域，"
        else:
            return ''

    def compare_grade_gen_text(self):
        """生成整体分析标题及一句分析文本。"""
        data = {}
        for grade in self.grade.grades:
            sch = self.grade_score[grade].avg
            last_year_grade_data = self.get_last_year_miss(grade=grade)
            year = last_year_grade_data['score']['total']
            data[grade] = sch - year
        total_dif = sum(data.values()) / len(data)
        diff = []
        if total_dif > 0:
            for k, v in data.items():
                if v < 0:
                    diff.append(k)
        elif total_dif < 0:
            for k, v in data.items():
                if v > 0:
                    diff.append(k)
        else:
            pass
        cond = self.compare_all_total
        if diff:
            grade_text = '除' + '、'.join([project.grade_simple[g] for g in diff]) + '年级外，'
        else:
            grade_text = ''
        title_text = f"{grade_text}各年级生命与健康素养{cond}全国平均水平"
        describe_text = f"{grade_text}各年级的健康素养水平{cond}{self.last_year_num}年的全国平均水平"
        return title_text, describe_text

    def describe_grade_text(self, category: str):
        """描述全年级对比情况
        :param category: total/gender  total：总体比全国 / gender：男生比女生。
        """
        bigger, smaller, others = [], [], []
        for grade in self.grade.grades:
            if category == 'total':
                first = self.grade_score[grade].avg
                last_year_grade_data = self.get_last_year_miss(grade=grade)
                second = last_year_grade_data['score']['total']
            else:
                first = self.summary_scores[grade]['M']
                second = self.summary_scores[grade]['F']
            if first - second >= 5:
                bigger.append(f"{project.grade_simple[grade]}年级")
            elif first - second <= -5:
                smaller.append(f"{project.grade_simple[grade]}年级")
            else:
                others.append(f"{project.grade_simple[grade]}年级")
        res = ''
        if bigger:
            local_codes = self._build_codes(bigger)
            res += f"{local_codes}"
            match category:
                case 'total':
                    res += '分数明显高于全国平均分数，'
                case 'gender':
                    res += '男生分数明显高于女生分数，'

        if smaller:
            local_codes = self._build_codes(smaller)
            res += f"{local_codes}"
            match category:
                case 'total':
                    res += "分数明显低于全国平均分数，"
                case 'gender':
                    res += "男生分数明显低于女生分数，"
        if (bigger or smaller) and ((len(bigger) + len(smaller)) < len(self.grade.grades)):
            res += f'其他年级分数没有明显差异。'
        if not bigger and not smaller:
            res = f'分数对比无明显差异。'
        return res

    @property
    def grade_class_student_table(self):
        scores = self.final_scores
        records = []
        grade_key, cls_key, gender_key = 'grade', 'cls', 'gender'
        for grade, grade_group in scores.groupby(grade_key):
            grade_base = {
                grade_key: grade,
                f"{grade_key}_count": len(grade_group),
                f"{grade_key}_{cls_key}_count": len(grade_group.groupby(cls_key)),
            }
            for cls, grade_cls_group in grade_group.groupby(cls_key):
                class_count = len(grade_cls_group)
                class_base = {
                    cls_key: cls,
                    f"{cls_key}_count": class_count,
                }
                for gender, grade_cls_gender_group in grade_cls_group.groupby(gender_key):
                    gender_count = len(grade_cls_gender_group)
                    class_base[f"{gender}_count"] = gender_count
                    class_base[f"{gender}_percent"] = round(gender_count / class_count * 100, 1)
                records.append(class_base | grade_base)
        records.sort(key=lambda x: (int(x['grade']) * 100 + int(x['cls'])))
        return records



from contextlib import contextmanager

import MySQLdb
import mongoengine
import sqlalchemy
from redis.client import StrictRedis

from hawa.base.decos import singleton
from hawa.config import project

metadata = sqlalchemy.MetaData()


@singleton
class DbUtil:
    _conn = None
    _cursor_conn = None
    _engine_conn = None

    @property
    def db_engine(self):
        database_url = f"{project.DB_MODE}://{project.DB_USER}:{project.DB_PSWD}@{project.DB_HOST}/" \
                       f"{project.DB_NAME}?charset=utf8"
        engine = sqlalchemy.create_engine(database_url, pool_pre_ping=True)
        return engine

    @contextmanager
    def engine_conn(self):
        self._engine_conn = self.db_engine.connect()
        yield self._engine_conn
        self._engine_conn.close()

    def connect(self):
        return MySQLdb.connect(
            host=project.DB_HOST,
            port=project.DB_PORT,
            user=project.DB_USER,
            passwd=project.DB_PSWD,
            db=project.DB_NAME,
        )

    @property
    def cursor(self):
        """only for test"""
        return self.cursor_conn.cursor(cursorclass=MySQLdb.cursors.DictCursor)

    @property
    def cursor_conn(self):
        if not self._cursor_conn:
            self._cursor_conn = self.connect()
        if not self._cursor_conn.open:
            self._cursor_conn = self.connect()
        # is self._cursor_conn alive?
        try:
            self._cursor_conn.ping()
        except Exception as e:
            project.logger.info(f"{type(e)=} {e=}")
            self._cursor_conn = self.connect()
        return self._cursor_conn

    def query_by_sql(self, sql: str, mode: str = 'all'):
        """
        :param sql:
        :param mode: all or one
        :return: list or dict by mode
        """
        with self.cursor_conn.cursor(cursorclass=MySQLdb.cursors.DictCursor) as cursor:
            cursor.execute(sql)
            match mode:
                case 'all':
                    return cursor.fetchall()
                case 'one':
                    return cursor.fetchone()


class MongoUtil:
    @classmethod
    def connect(
            cls, db: str = '', host: str = '', port: int = 0,
            username: str = '', pswd: str = '', source: str = ''
    ):
        mongoengine.connect(
            db or project.MONGO_DB,
            host=host or project.MONGO_HOST, port=port or project.MONGO_PORT,
            username=username or project.MONGO_USER,
            password=pswd or project.MONGO_PSWD,
            authentication_source=source or project.MONGO_AUTH_DB
        )


@singleton
class RedisUtil:
    @property
    def conn(self):
        return StrictRedis(
            host=project.REDIS_HOST,
            db=project.REDIS_DB,
            decode_responses=True,
        )
